# common test fixtures

import asyncio
import json
import logging
import multiprocessing as mp
import os
import queue
from pathlib import Path
from subprocess import run
from time import sleep, time

import MoralisSDK.api
import psutil
import pytest
from ape_apeman.context import APE
from eth_hash.auto import keccak
from eth_utils import to_checksum_address, to_hex

from moralis_streams_client import MoralisStreamsApi, defaults, server
from moralis_streams_client.signature import Signature
from moralis_streams_client.tunnel import NgrokTunnel
from moralis_streams_client.webhook import Webhook

info = logging.info

ECOSYSTEM = "ethereum"
NETWORK = "goerli"
PROVIDER = "alchemy"
CHAIN_ID = "0x5"


@pytest.fixture(scope="session", autouse=True)
def assert_startup_procs():
    assert (
        run(["pgrep", "ngrok"]).returncode != 0
    ), "ngrok processs running before test"
    assert (
        run(["pgrep", "webhook-server"]).returncode != 0
    ), "webhook process running before test"
    try:
        yield True
    finally:
        pass
        # run(["pkill", "ngrok"], check=False)
        # run(["pkill", "msc"], check=False)


@pytest.fixture(scope="session")
def event_loop():
    policy = asyncio.get_event_loop_policy()
    loop = policy.new_event_loop()
    yield loop
    loop.close()


@pytest.fixture
def event_keys():
    return ["body", "headers", "id", "method", "path", "relay"]


@pytest.fixture
def config():
    def _config(key):
        return os.environ[key]

    return _config


@pytest.fixture
def moralis(config):
    api = MoralisSDK.api.MoralisApi()
    api.set_api_key(config("MORALIS_API_KEY"))
    return api


@pytest.fixture(scope="module")
def webhook():
    return Webhook()


@pytest.fixture
def ecosystem():
    return ECOSYSTEM


@pytest.fixture
def network():
    return NETWORK


@pytest.fixture
def provider():
    return PROVIDER


@pytest.fixture
def chain_id():
    return CHAIN_ID


@pytest.fixture
def streams(api_key, api_url):
    return MoralisStreamsApi(api_key=api_key, url=api_url)


@pytest.fixture
def ape(ecosystem, network, provider):
    with APE(ecosystem=ecosystem, network=network, provider=provider) as ape:
        yield ape


@pytest.fixture
def system_address(config):
    address = config("TEST_SYSTEM_ADDRESS")
    return to_checksum_address(address)


@pytest.fixture
def system_key(config):
    return config("TEST_SYSTEM_KEY")


@pytest.fixture
def user_address(config):
    address = config("TEST_USER_ADDRESS")
    return to_checksum_address(address)


@pytest.fixture
def user_key(config):
    return config("TEST_USER_KEY")


@pytest.fixture
def api_key(config):
    return config("MORALIS_API_KEY")


@pytest.fixture
def api_url():
    return defaults.STREAMS_URL


@pytest.fixture
def server_addr():
    return defaults.SERVER_ADDR


@pytest.fixture
def server_port():
    return defaults.SERVER_PORT


@pytest.fixture
def ethersieve_contract_address(config):
    return config("ETHERSIEVE_CONTRACT_ADDRESS")


@pytest.fixture
def background_contract_address():
    return config("BACKGROUND_CONTRACT_ADDRESS")


def is_ok(response):
    return not response.is_error


def wait_for_it(address, port, timeout=15):
    time_expired = time() + timeout
    while (
        run(["nc", "-z", str(address), str(port)], check=False).returncode != 0
    ):
        assert time() < time_expired
        sleep(1)


class WebhookServerProcess:
    def __init__(
        self,
        addr=defaults.SERVER_ADDR,
        port=defaults.SERVER_PORT,
        tunnel=True,
        enable_buffer=True,
        debug=True,
        log_file="webhook.log",
    ):
        print("starting webhook_server...")
        self.addr = addr
        self.port = port
        self.log_file = log_file
        self.webhook = Webhook(
            addr=addr,
            port=port,
            tunnel=tunnel,
            debug=debug,
            enable_buffer=enable_buffer,
        )

    async def __aenter__(self):
        await self.webhook.start(
            wait=True, log_file=str(Path(".") / self.log_file)
        )
        wait_for_it(self.addr, self.port)
        assert await self.webhook.clear()
        info(await self.webhook.tunnel_url())
        procs = await self.webhook.processes()
        for proc in procs:
            info(proc)
        return self.webhook

    async def __aexit__(self, _type, exc, tb):
        def webhook_exit(proc):
            print(f"webhook_server {proc} terminated")

        print("\nstopping webhook_server processes...")
        await self.webhook.stop(wait=True, callback=webhook_exit)


@pytest.fixture
async def webhook_process():
    async with WebhookServerProcess() as webhook:
        yield webhook


@pytest.fixture
def webhook_process_class():
    return WebhookServerProcess


@pytest.fixture
def dump():
    def _dump(obj):
        info(json.dumps(obj, indent=2))

    return _dump


@pytest.fixture
async def webhook_tunnel_url(webhook):
    url = await webhook.tunnel_url()
    assert isinstance(url, str)
    assert url.startswith("http://")
    return url



import collections
import logging

import httpx

from . import defaults, settings

logger = logging.getLogger(__name__)
debug = logger.debug
logger.setLevel(settings.LOG_LEVEL)


class EventQueue:
    def __init__(self):
        debug("init")
        self.events = collections.deque([])
        self.buffer_enabled = settings.BUFFER_ENABLE
        self.relay_url = settings.RELAY_URL
        self.relay_header = settings.RELAY_HEADER
        self.relay_key = str(settings.RELAY_KEY)
        self.relay_id_header = settings.RELAY_ID_HEADER

    async def append(self, event):
        debug("append")
        if self.relay_url:
            _response = await self.forward(event)
            event.relay = dict(
                url=_response.url,
                status_code=_response.status_code,
                text=_response.text,
                headers=dict(_response.headers),
            )
        else:
            event.relay = None
        if self.buffer_enabled:
            self.events.append(event)
        return event.id

    async def forward(self, event):
        debug("forward")
        headers = event.headers
        if self.relay_header and self.relay_key:
            headers[self.relay_header] = self.relay_key
        headers[self.relay_id_header] = str(event.id)
        debug(f"post({self.relay_url} {headers} {event.body})")
        async with httpx.AsyncClient() as client:
            ret = await client.post(
                self.relay_url, headers=headers, json=event.body
            )
        debug(f"ret={ret}")
        return ret

    async def list(self):
        # create a list from the event_queue, preserving order
        debug("list")
        ret = []
        tq = self.events.copy()
        try:
            while True:
                ret.append(tq.popleft())
        except IndexError:
            pass
        return ret

    async def clear(self):
        debug("clear")
        self.events.clear()
        return "cleared"

    async def lookup(self, event_id, delete):
        debug("lookup")
        tq = self.events.copy()
        if delete:
            self.events.clear()
        found = None
        while True:
            try:
                event = tq.popleft()
                if event.id == event_id:
                    found = event
                    if delete is False:
                        break
                elif delete is True:
                    self.events.append(event)
            except IndexError:
                break

        return found



# cli tests

import logging
import os
import shlex
import subprocess

import pytest

from moralis_streams_client.webhook import Webhook

info = logging.info


@pytest.fixture
def run():
    def _run(cmd, check=True, expect_error=False):
        if isinstance(cmd, list):
            pass
        else:
            cmd = shlex.split(cmd)
        proc = subprocess.run(cmd, check=False, capture_output=True, text=True)
        if check:
            if expect_error:
                assert proc.returncode != 0, proc
            else:
                assert proc.returncode == 0, proc
        return proc

    return _run


def as_lines(string):
    return string.strip().split("\n")


def test_cli_no_args(run):
    result = run("msc")
    assert result.stdout
    info(result.stdout)
    assert result.stdout.startswith("Usage:")


def test_cli_help(run):
    result = run("msc", "--help")
    assert result.stdout
    info(result.stdout)


async def test_cli_start_stop(run):
    run("webhook -p 8082 ps", expect_error=True)
    run("webhook -p 8082 -d -l DEBUG -T start")
    run("webhook -p 8082 ps")
    ret = run("webhook -p 8082 hello")
    assert "Hello, Webhook!" in ret.stdout
    await Webhook(port=8082).stop()
    # run("webhook -p 8082 stop")
    run("webhook -p 8082 ps", expect_error=True)


async def test_cli_two_start_stop(run):
    run("webhook -d -l DEBUG -T start")
    run("webhook -p 8081 -d -l DEBUG -T start")

    ret = run("webhook ps")
    lines = as_lines(ret.stdout)
    [info(line) for line in lines]
    assert len(lines) == 1

    ret = run("webhook -p 8081 ps")
    lines = as_lines(ret.stdout)
    [info(line) for line in lines]
    assert len(lines) == 1

    run("webhook ps")
    await Webhook().stop()
    run("webhook ps", expect_error=True)

    run("webhook -p 8081 ps")
    await Webhook(port=8081).stop()
    run("webhook -p 8081 ps", expect_error=True)


async def test_cli_debug_log(run, shared_datadir):
    logfile = shared_datadir / "webhook_test.log"
    os.environ["WEBHOOK_LOG_LEVEL"] = "DEBUG"
    os.environ["WEBHOOK_DEBUG"] = "1"
    os.environ["WEBHOOK_LOG_FILE"] = str(logfile)
    run("webhook ps", expect_error=True)
    run("webhook -T start")
    run("webhook ps")
    assert logfile.is_file()
    assert logfile.stat().st_size != 0
    await Webhook().stop()



# test relay mode

import atexit
import time
from logging import critical, info
from pathlib import Path
from subprocess import run
from uuid import uuid4

import pytest
from httpx import Headers

from moralis_streams_client import settings
from moralis_streams_client.app import obscure_key
from moralis_streams_client.webhook import Webhook

from .conftest import WebhookServerProcess, wait_for_it

EVENT_COUNT = 10

TEST_RELAY_KEY = "test_relay_key"
RELAY_KEY_HEADER = "X-API-Key"
RELAY_PORT = 8080
TARGET_PORT = 8081


@pytest.fixture(scope="module", autouse=True)
async def webhook_process():
    async with WebhookServerProcess() as webhook:
        yield webhook


@pytest.fixture
def relay_port():
    return RELAY_PORT


@pytest.fixture
def target_port():
    return TARGET_PORT


@pytest.fixture(scope="module")
async def verify_webhook():
    async def _verify_webhook(_webhook, port):
        assert await _webhook.server_running() is True
        procs = await _webhook.processes()
        assert isinstance(procs, list)
        assert len(procs) > 0
        assert _webhook.port == port
        assert await _webhook.clear()
        return True

    return _verify_webhook


@pytest.fixture
def relay_url(target_port):
    return f"http://localhost:{target_port}/contract/event"


@pytest.fixture
def relay_key():
    return TEST_RELAY_KEY


@pytest.fixture
def relay_header():
    return RELAY_KEY_HEADER


@pytest.fixture(scope="module")
async def relay(webhook, verify_webhook):
    """configue running webhook to relay messages to target webhook"""
    relay_port = RELAY_PORT
    target_port = TARGET_PORT
    relay_key = TEST_RELAY_KEY
    relay_header = RELAY_KEY_HEADER
    relay_url = f"http://localhost:{target_port}/contract/event"

    _relay = webhook
    await verify_webhook(_relay, relay_port)
    set_result = await _relay.relay(
        url=relay_url, key=relay_key, header=relay_header
    )
    assert isinstance(set_result, dict)
    get_result = await _relay.relay()
    assert get_result == dict(
        url=relay_url, header=relay_header, key=obscure_key(relay_key)
    )
    try:
        yield _relay
    finally:
        assert await _relay.events() == []
        await _relay.relay(enable=False)
        assert await _relay.relay() == dict(url=None, header=None, key=None)
        assert await _relay.buffer() is True


@pytest.fixture(scope="module")
async def target(verify_webhook):
    target_port = TARGET_PORT
    _target = Webhook(
        debug=True,
        port=target_port,
        tunnel=False,
    )
    log_file = Path(".") / "relay.log"
    await _target.start(wait=True, log_file=str(log_file))
    wait_for_it(_target.addr, _target.port)
    assert await _target.server_running()
    assert await verify_webhook(_target, target_port)
    await _target.relay(enable=False)
    try:
        yield _target
    finally:
        assert await _target.events() == []
        await _target.relay(enable=False)
        assert await _target.relay() == dict(url=None, header=None, key=None)
        assert await _target.buffer() is True
        await _target.stop()


async def test_relay_messages(
    relay, target, relay_url, relay_key, relay_header
):
    assert await relay.clear()
    assert await target.clear()

    assert await relay.relay() == dict(
        url=f"http://localhost:{target.port}/contract/event",
        header=relay_header,
        key=obscure_key(relay_key),
    )
    assert await target.relay() == dict(url=None, header=None, key=None)

    assert len(await target.events()) == 0
    assert len(await relay.events()) == 0

    sent_events = {}
    for i in range(EVENT_COUNT):
        event = dict(test=f"payload_{i}", count=i, id=str(uuid4()))
        event_id = await relay.inject(event)
        sent_events[event_id] = event

    while len(await target.events()) < EVENT_COUNT:
        info(f"target_events_length={target.events()}")
        time.sleep(1)

    target_events = await target.events()

    sent_ids = list(sent_events.keys())
    target_ids = [t["id"] for t in target_events]
    assert len(sent_ids) == len(target_ids)

    for i, sid in enumerate(sent_events.keys()):
        sbody = sent_events[sid]
        tev = target_events[i]
        tbody = tev["body"]
        assert sbody == tbody
        assert isinstance(tev["headers"], dict)
        assert relay_header in Headers(tev["headers"])
        assert Headers(tev["headers"])[relay_header] == relay_key

    for event in target_events:
        event_id = Headers(event["headers"])["x-relay-id"]
        assert await relay.event(event_id)
        await relay.delete(event_id)

    assert await relay.events() == []


async def test_relay_id(
    relay, target, relay_url, relay_key, relay_header, dump
):
    assert await relay.clear()
    assert await target.clear()

    msg_id = str(uuid4())
    event = dict(test="payload", id=msg_id)
    await relay.inject(event)
    tevs = await target.events()
    assert len(tevs) == 1
    dump(tevs[0])
    assert tevs[0]["body"]["id"] == msg_id
    await relay.clear()
    await target.clear()


def _make_huge_event(size):
    msg_id = str(uuid4())
    huge = "a" + "u" * size + "gh"
    event = dict(test="huge", id=msg_id, castle_of=huge)
    return event


async def test_relay_huge(
    relay, target, relay_url, relay_key, relay_header, dump
):
    assert await relay.clear()
    assert await target.clear()

    event = _make_huge_event(8_000_000)

    await relay.inject(event)

    tevs = await target.events()
    assert len(tevs) == 1
    assert tevs[0]["body"]["id"] == event["id"]
    assert len(tevs[0]["body"]["castle_of"]) == len(event["castle_of"])
    await relay.clear()
    await target.clear()


async def test_relay_recover_huge(
    relay, target, relay_url, relay_key, relay_header, dump
):
    assert await relay.clear()
    assert await target.clear()

    event = _make_huge_event(11_000_000)

    with pytest.raises(Exception) as exc:
        await relay.inject(event)
    info(f"Exception: {exc}")

    event = _make_huge_event(9_000_000)
    await relay.inject(event)

    tevs = await target.events()
    assert len(tevs) == 1
    assert tevs[0]["body"]["id"] == event["id"]
    assert len(tevs[0]["body"]["castle_of"]) == len(event["castle_of"])
    await relay.clear()
    await target.clear()



# webhook server process

import json
import logging
import os
import re
import subprocess
import sys
import time
from pathlib import Path

import httpx
import psutil

from . import settings
from .logconfig import configure_logging
from .signature import Signature

logger = logging.getLogger(__name__)
debug = logger.debug

PROCESS_START_TIMEOUT = 5
PROCESS_STOP_TIMEOUT = 10


class Webhook:
    def __init__(
        self,
        *,
        debug=None,
        addr=None,
        port=None,
        base_url=None,
        tunnel=None,
        relay_url=None,
        relay_key=None,
        relay_header=None,
        enable_buffer=None,
        moralis_api_key=None,
        log_level=None,
        log_file=None,
    ):
        self.debug = settings.DEBUG if debug is None else debug
        self.addr = addr or settings.ADDR
        self.port = port or settings.PORT
        self.tunnel = settings.TUNNEL if tunnel is None else tunnel
        self.relay_url = relay_url or settings.RELAY_URL
        self.relay_key = relay_key or settings.RELAY_KEY
        self.relay_header = relay_header or settings.RELAY_HEADER
        self.enable_buffer = (
            settings.BUFFER_ENABLE if enable_buffer is None else enable_buffer
        )
        self.moralis_api_key = moralis_api_key or settings.MORALIS_API_KEY
        self.log_level = log_level or settings.LOG_LEVEL
        self.log_file = log_file or settings.LOG_FILE
        base_url = base_url or f"http://{self.addr}:{self.port}/"
        self.base_url = base_url.strip("/") + "/"
        self.signature = Signature()
        self.proc = None
        configure_logging()

    async def _request(self, method, path, **kwargs):
        # generate a signature checksum
        raise_for_status = kwargs.pop("raise_for_status", True)
        kwargs.setdefault("json", None)
        kwargs.setdefault("headers", {})
        if kwargs["json"] is None:
            body = b""
        else:
            body = kwargs["json"]
        kwargs["headers"].update(self.signature.headers(body))

        async with httpx.AsyncClient() as client:
            self.response = await client.request(
                method, self.base_url + path, **kwargs
            )
        if raise_for_status:
            self.response.raise_for_status()
        body = self.response.json()
        return body["result"]

    async def hello(self):
        """send and receive a friendly greeting"""
        return await self._request("GET", "hello")

    async def clear(self):
        """clear the event buffer"""
        return await self._request("GET", "clear")

    async def tunnel_url(self):
        """return the tunnel url"""
        return await self._request("GET", "tunnel")

    async def buffer(self, enable=None):
        """set buffer enable"""
        if enable is None:
            method = "GET"
            args = {}
        else:
            method = "POST"
            args = dict(enable=enable)

        return await self._request(method, "buffer", json=args)

    async def relay(self, url=None, key=None, header=None, enable=None):
        """update relay configuraton"""
        args = {}
        method = "GET"
        if enable is None and url is not None:
            enable = True
        if enable is True:
            method = "POST"
            args["url"] = url
            args["key"] = key
            args["header"] = header
        elif enable is False:
            method = "POST"
            args["url"] = None
            args["key"] = None
            args["header"] = None

        return await self._request(method, "relay", json=args)

    async def inject(self, event):
        """process an event as a received callback"""
        return await self._request("POST", "contract/event", json=event)

    async def event(self, event_id):
        """return event by id"""
        return await self._request("GET", f"event/{event_id}")

    async def delete(self, event_id):
        """delete event by id"""
        return await self._request("DELETE", f"event/{event_id}")

    async def events(self):
        """return a list of all events"""
        return await self._request("GET", "events")

    async def shutdown(self):
        """request server shutdown"""
        return await self._request("GET", "shutdown")

    async def start(self, wait=True, log_file=None):
        """start a server process"""
        env = os.environ.copy()

        if log_file:
            self.log_file = log_file

        env["WEBHOOK_ADDR"] = self.addr
        env["WEBHOOK_PORT"] = str(self.port)
        env["WEBHOOK_LOG_FILE"] = str(self.log_file)
        env["WEBHOOK_LOG_LEVEL"] = str(self.log_level)
        env["WEBHOOK_DEBUG"] = "1" if self.debug else "0"
        env["WEBHOOK_TUNNEL"] = "1" if self.tunnel else "0"
        env["WEBHOOK_BUFFER_ENABLE"] = "1" if self.enable_buffer else "0"
        if self.relay_url:
            env["WEBHOOOK_RELAY_URL"] = self.relay_url
        if self.relay_header:
            env["WEBHOOOK_RELAY_HEADER"] = self.relay_header
        if self.relay_key:
            env["WEBHOOK_RELAY_KEY"] = str(self.relay_key)
        if self.moralis_api_key:
            env["WEBHOOK_API_KEY"] = str(self.moralis_api_key)

        cmd = ["webhook-server", "--port", str(self.port)]
        debug(f"{cmd=}")
        for k, v in env.items():
            if k.startswith("WEBHOOK"):
                debug(f"  {k}={v}")

        pid = subprocess.Popen(
            cmd,
            env=env,
            stderr=subprocess.STDOUT,
            stdout=subprocess.DEVNULL,
            close_fds=True,
        ).pid
        debug(f"{pid=}")

        if wait:
            timeout = time.time() + PROCESS_START_TIMEOUT
            while await self.server_running() is False:
                time.sleep(0.25)
                if time.time() > timeout:
                    raise TimeoutError
        return pid

    async def processes(self):
        """return list of server processess"""
        pattern = f".*webhook.*--port {self.port}.*"
        procs = []
        for p in psutil.process_iter():
            cmdline = " ".join(p.cmdline())
            if re.match(pattern, cmdline):
                debug(f"process: {p}")
                procs.append(p)
        return procs

    async def server_running(self):
        """return bool indicating if server is running"""
        return bool(await self.processes())

    async def stop(self, wait=True, callback=None):
        """signal running server processes to terminate"""
        if await self.server_running():
            procs = await self.processes()
            for p in procs:
                debug(f"terminate: {p}")
                p.terminate()

            if wait:
                gone, alive = psutil.wait_procs(
                    await self.processes(),
                    timeout=PROCESS_STOP_TIMEOUT,
                    callback=callback,
                )
                for p in alive:
                    p.kill()

                if await self.server_running():
                    await self.shutdown()

            timeout = time.time() + PROCESS_STOP_TIMEOUT
            while wait and await self.server_running():
                time.sleep(0.25)
                if time.time() > timeout:
                    raise TimeoutError



# logging configuration

import logging
import sys

from . import settings


def configure_logging(log_level=None, log_file=None, log_format=None):

    log_level = log_level or settings.LOG_LEVEL
    log_file = log_file or settings.LOG_FILE
    log_format = log_format or settings.LOG_FORMAT

    args = dict(force=True, level=log_level, format=log_format)
    if log_file:
        args["filename"] = str(log_file)
    else:
        args["stream"] = sys.stderr

    logging.basicConfig(**args)

    logging.getLogger("urllib3.connectionpool").setLevel(logging.WARNING)



#!/usr/bin/env python3

import logging
import sys

import click
import uvicorn

from moralis_streams_client import settings
from moralis_streams_client.app import app
from moralis_streams_client.logconfig import configure_logging
from moralis_streams_client.tunnel import NgrokTunnel


class ServerProcess:
    def __init__(self, **kwargs):

        self.debug = kwargs.get("debug", settings.DEBUG)
        self.addr = kwargs.get("addr", settings.ADDR)
        self.port = kwargs.get("port", settings.PORT)
        self.tunnel = kwargs.get("tunnel", settings.TUNNEL)
        self.log_level = kwargs.get("log_level", settings.LOG_LEVEL)

        self.config = kwargs.get("config", {})
        self.config["debug"] = settings.DEBUG
        self.config["log_level"] = settings.LOG_LEVEL
        app.state.tunnel_url = None
        app.state.config = self.config

        configure_logging()

        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(settings.LOG_LEVEL)
        self.info = self.logger.info

    def run(self):
        def fileConfig(*args, **kwargs):
            self.info("uvicorn logging config disabled")

        uvicorn.config.logging.config.fileConfig = fileConfig

        def _uvicorn_run():
            return uvicorn.run(
                "moralis_streams_client.app:app",
                host=self.addr,
                port=self.port,
                log_level=self.log_level.lower(),
                log_config="disable",
            )

        if self.tunnel:
            self.info("starting ngrok tunnel...")
            with NgrokTunnel(
                self.port, str(settings.NGROK_AUTHTOKEN)
            ) as tunnel:
                app.state.tunnel_url = tunnel.public_url
                return _uvicorn_run()
        else:
            self.info("ngrok tunnel disabled")
            return _uvicorn_run()


@click.command
@click.option(
    "-p",
    "--port",
    type=int,
    required=True,
    help="server listen port",
)
def server(port):
    sys.exit(ServerProcess(port=port).run())
